from typing import Dict, List, NamedTuple, Optional, Type


class Error(NamedTuple):
    desc: str
    help_text: Optional[str] = None


class ModCFGError(Exception):
    """Base error class for ModCFG."""


class ModCFGSyntaxError(ModCFGError):
    """There was an error parsing the ModCFG document"""


class InvalidEscape(ModCFGSyntaxError):
    """You used an invalid escape sequence"""


class InvalidDateFormat(ModCFGSyntaxError):
    """Your date(time)s cannot be parsed"""

    label: Error = Error(
        "Your date(time)s cannot be parsed",
        "Remember, it's ISO 8601 format (TODO: Full ISO 8601 support). (YYYY-MM-DD for dates,\n"
        "YYYY-MM-DD[*HH[:MM[:SS[.fff[fff]]]][+HH:MM[:SS[.ffffff]]]] for datetimes)",
    )


class DumpError(ModCFGError):
    """Failed to dump"""


class EnumResolveError(ModCFGError):
    """There was an error when resolving enumeration(s)"""


class EmptyBody(ModCFGSyntaxError):
    label: Error = Error(
        "Your module body is empty", "Try adding stuff to the body of your module"
    )


class BadIndentation(ModCFGSyntaxError):
    label: Error = Error(
        "Your indentaton is inconsistent",
        "Did you mean to make the indentation the same?",
    )


class Y_U_NO_TEXT(ModCFGSyntaxError):
    label: Error = Error("There is no content to parse", "Try adding some content")


class BadName(ModCFGSyntaxError):
    label: Error = Error(
        "Your name is not valid", "Just so you know, it should not start with a number"
    )


class BadKeyChar(ModCFGSyntaxError):
    label: Error = Error(
        "The keys are not valid",
        "That there are no newlines before and after it",
    )


class InvalidKeyChar(ModCFGSyntaxError):
    label: Error = Error(
        "The keys are not valid",
        "Make sure the keys are either one of the following characters:\n - {}".format(
            "\n - ".join(map(repr, (":", "=>", "=", "->")))
        ),
    )


class MixedModuleContents(ModCFGSyntaxError):
    label: Error = Error(
        "The keys are not valid",
        "You can only put a dict or a list in a module, mutually exclusive",
    )


ERROR_MAP: Dict[Type[ModCFGSyntaxError], List[str]] = {
    EmptyBody: ["""mod a:"""],
    BadIndentation: ["""mod s:\n - e\n  - e"""],
    Y_U_NO_TEXT: ["", "\n", "\n   \n"],
    BadName: [
        "mod s:\n 1 = 1\n",
        "mod s:\n e = e\n 1 = 1\n",
        "mod s:\n w = null #eeeee\n 1 = 1\n",
        "mod a:\n b = c #eeeee\n 1 = 1\n",
    ],
    InvalidKeyChar: ["mod a:\n    b * c"],
    MixedModuleContents: [
        "mod a:\n    a=b\n     - b",
        "mod a:\n    a = b\n    - c\n    b = b",
    ],
    InvalidDateFormat: ["{main: date(a)}", "{main:datetime(b)}"],
}



# TODO: Support for single values

import datetime
import enum
import textwrap
from typing import Any, Dict, List, Optional, Type, TypeVar, Union

import lark
import lark.indenter

from . import components, errors, lark_components, utils

# TODO: Better type hints


def loads(
    s: str,
    enums: Optional[List[enum.Enum]] = None,
    enum_resolve_fail_silently: bool = False,
    enum_ambiguity_check: bool = True,
    inline: bool = False,
) -> Union[
    Union[components.Module, List[Any], Dict[str, Any]],
    List[Union[components.Module, List[Any], Dict[str, Any]]],
]:
    """Serializes the document, `s`, to Python objects.

    Args:
        s (str): The ModCFG document to parse and evaluate.
        enums (Optional[List[enum.Enum]]): The list of Python enums that should be serialized from in the document. Defaults to None.
        enum_resolve_fail_silently (bool): Implicitly convert enum values to `str` if couldn't resolve the enum. Defaults to True.
        enum_ambiguity_check (bool): Fail if the enums given are ambigous. Defaults to True.
        inline (bool): Inline the output. See docs for more details. Defaults to False.

    Returns:
        Union[Union[components.Module, List[Any], Dict[str, Any]],List[Union[components.Module, List[Any], Dict[str, Any]]]]: The serialized Python object.

    Raises:
        `errors.ModCFGError`: There was an error processing your document

    :canonical: modcfg.parser
    """
    # We must do the tree separately because we're also parsing indentation
    # See https://github.com/lark-parser/lark/issues/818
    transformer = lark_components.MainTransformer(
        enums=enums,
        enum_resolve_fail_silently=enum_resolve_fail_silently,
        inline=inline,
        enum_ambiguity_check=enum_ambiguity_check,
    )
    try:
        tree = lark_components.grammar.parse(s)
    except lark.exceptions.UnexpectedInput as parse_error:
        error: Type[errors.ModCFGSyntaxError] = parse_error.match_examples(
            lark_components.grammar.parse, errors.ERROR_MAP, use_accepts=True
        )
        if error is None:  # pragma: no cover
            raise
        raise error(
            parse_error.get_context(s),
            parse_error.line,  # pylint: disable=E1101
            parse_error.column,  # pylint: disable=E1101
        ) from parse_error
    else:
        try:
            return transformer.transform(tree)
        except lark.exceptions.VisitError as error:
            raise error.orig_exc


T = TypeVar("T")


def _unparse(
    thing: Union[
        str,
        List[T],
        Dict[str, T],
        datetime.date,
        datetime.datetime,
        enum.Enum,
        None,
        bool,
    ],
    # enums: Optional[List[enum.Enum]] = None,
) -> str:
    if isinstance(thing, bool):
        return "true" if thing else "false"
    if isinstance(thing, (str, float, int)):
        return repr(thing)
    if isinstance(thing, list):
        return "[" + ", ".join(_unparse(k) for k in thing) + "]"
    if isinstance(thing, dict):
        return (
            "{"
            + ", ".join(f"{_unparse(k)}: {_unparse(v)}" for k, v in thing.items())
            + "}"
        )
    if isinstance(thing, (datetime.date, datetime.datetime)):
        return f"{thing.__class__.__name__}({thing})\n"
    if thing is None:
        return "null"
    if isinstance(thing, enum.Enum):
        return ":" + thing.__class__.__name__ + "." + thing.name
    raise errors.DumpError(f"Failed to dump object {type(thing).__name__}")


def _dump_module(module: components.Module) -> str:
    # if len(module.contents) == 0:
    #     raise errors.EmptyBody(
    #         f"components.Module must contain something (module {module.name!r} of arg `module`)"
    #     )
    output = f"mod {module.name}:\n"
    inner = ""
    inner += textwrap.indent(_unparse(module.contents).strip("\n"), utils.INDENT)
    return output + inner + "\n"


def dumps(
    objects: Union[
        List[T],
        Dict[str, T],
        components.Module,
    ],
) -> str:
    R"""Converts objects to `str` that can be `loads`\ ed back.

    Basically `json.dumps` but for ModCFG.

    Args:
        objects (Union[List[Any], Dict[str, Any], components.Module]): The objects to dump.

    Returns:
        str: Returns the dumped `str`.

    Raises:
        `errors.DumpError`: Invalid object (couldn't dump it)

    :canonical: modcfg.parser
    """
    output = ""
    if isinstance(objects, components.Module):  # Single module
        return _dump_module(objects)
    if isinstance(objects, (list, dict)):
        # if not objects:
        #     raise ValueError("Dict must contain something.")
        for obj in objects:
            if isinstance(obj, components.Module):
                output += _dump_module(obj)
            else:
                output += _unparse(obj)
    else:
        # raise ValueError("Must be a list or dict of stuff")
        output += _unparse(objects)
    return textwrap.dedent(output.strip("\n"))



import ast
import datetime
import enum
import textwrap
from pathlib import Path
from typing import Any, Dict, List, Union

import lark.indenter

from . import components, errors, utils

here: Path = Path(__file__).parent


class Indenter(lark.indenter.Indenter):
    NL_type = "_NL"
    OPEN_PAREN_types: List[str] = ["_OPEN_BRACKET", "_OPEN_BRACE"]
    CLOSE_PAREN_types: List[str] = ["_CLOSE_BRACKET", "_CLOSE_BRACE"]
    INDENT_type = "_INDENT"
    DEDENT_type = "_DEDENT"
    tab_len = 4


class MainTransformer(lark.Transformer):
    # pylint: disable=R0201,C0116,C0103
    ALL_NUMBERS = lambda self, x: ast.literal_eval(x)  # type: ignore
    CNAME = lambda self, x: str(x)  # type: ignore
    # TODO: dateutil for full iso support
    def DATE(self, x: str) -> datetime.date:
        clean = utils.FUNCTIONS["date"].search(x).group(1)
        try:
            return datetime.date.fromisoformat(clean)
        except ValueError as error:
            raise errors.InvalidDateFormat("Invalid date format: %s" % clean) from error

    def DATETIME(self, x: str) -> datetime.datetime:
        clean = utils.FUNCTIONS["datetime"].search(x).group(1)
        try:
            return datetime.datetime.fromisoformat(clean)
        except ValueError as error:
            raise errors.InvalidDateFormat("Invalid date format: %s" % clean) from error

    # DATE = lambda self, x: datetime.date.fromisoformat(x.strip()[5:-1])
    # DATETIME = lambda self, x: datetime.datetime.fromisoformat(x[9:-1])
    dict = dict
    list = list
    NULL = lambda _, __: None  # type: ignore
    BOOLEAN = lambda _, thing: str(thing).strip() == "true"  # type: ignore
    pair = tuple

    def start(self, stuff: List[str]) -> Any:
        if self._options["inline"] and len(stuff) == 1:  # Inlining
            return stuff[0]
        return stuff

    def __init__(self, *args, **kwargs) -> None:
        self._options: Dict[str, Union[List[enum.Enum], bool]] = kwargs
        super().__init__(*args)

    def MULTILINE_STR(self, text: str) -> str:
        return utils.parse_strings(3, text)

    def STRING(self, text: str) -> str:
        return utils.parse_strings(1, text)

    def module(self, stuff) -> components.Module:  # type: ignore
        args = stuff

        return components.Module(
            name=args[0],
            contents=args[1],
        )  # TODO: Use dicts instead

    def ENUM(self, given: str) -> Union[str, enum.Enum]:  # TODO: Add abiguity finding
        # TODO: Refactor this
        if not self._options["enums"]:
            if self._options["enum_resolve_fail_silently"] is True:
                return str(given)
            raise errors.EnumResolveError(f"Could not resolve enum for {given}")
        thing = str(given[1:])
        parsed = thing.split(".", 1)
        if len(parsed) == 1:  # No enum name specifier
            assert thing == parsed[0]
            possibles = []
            for entry in self._options["enums"]:
                vmap = dict(entry.__members__)
                if thing in vmap:
                    if self._options["enum_ambiguity_check"] is False:
                        return vmap[thing]  # Short circut
                    possibles.append(vmap.pop(thing))
            if len(possibles) > 1:
                raise errors.EnumResolveError(
                    "Ambigous enumerations: {}".format(", ".join(map(repr, possibles)))
                )
            if len(possibles) == 0:
                if self._options["enum_resolve_fail_silently"] is True:
                    return str(given)
                raise errors.EnumResolveError(f"Could not resolve enum for {given}")
            assert len(possibles) == 1
            return possibles[0]
        assert len(parsed) == 2
        for entry in self._options["enums"]:
            if parsed[0] == entry.__name__:
                vmap = dict(entry.__members__)
                if parsed[1] in vmap:
                    return vmap[parsed[1]]
        if self._options["enum_resolve_fail_silently"] is True:
            return str(given)
        raise errors.EnumResolveError(f"Could not resolve enum for {given}")


grammar = lark.Lark(
    here.joinpath("grammar.lark").read_text(),
    postlex=Indenter(),
    parser="lalr",
    maybe_placeholders=True,
)



import re
import textwrap
from typing import Dict, Match, Pattern

# from . import errors

_unescape_map: Dict[str, str] = {
    '"': '"',
    "'": "'",
    "/": "/",
    "\\": "\\",
    "b": "\b",
    "f": "\f",
    "n": "\n",
    "r": "\r",
    "t": "\t",
}

_unescape_re: Pattern[str] = re.compile(
    r"\\"  # Non-escaped Escaping Backslash
    + r"(["
    + "".join(map(re.escape, _unescape_map))  # Anything in _unescape_map
    + r"]|(?:u[0-9A-Fa-f]{4})|(?:U[0-9A-Fa-f]{8})|(?:x[0-9A-Fa-f]{2}))"  # Or unicode
)


def _unescape(match: Match[str]) -> str:
    char = match.group(1)
    if char[0] in {"u", "U", "x"}:  # Unicode
        return chr(int(char[1:], 16))
    return _unescape_map[char]


def parse_escape(s: str) -> str:
    return _unescape_re.sub(_unescape, s)


INDENT: str = " " * 4


def _function(name: str) -> Pattern[str]:
    return re.compile(name + r"\s*\(\s*(.+)\s*\)")


FUNCTIONS: Dict[str, Pattern[str]] = {
    "datetime": _function("datetime"),
    "date": _function("date"),
}


def parse_strings(str_char_len: int, text: str) -> str:
    actions = []
    output = text[:-str_char_len]

    if output.startswith("t"):
        output = output[1:]
        assert not output.startswith("t")
        actions.append(textwrap.dedent)
    if output.startswith("l"):
        output = output[1:]
        assert not output.startswith("l")
        actions.append(lambda x: x.lstrip())
    if output.startswith("r"):
        output = output[1:]
        assert not output.startswith("r")
        actions.append(lambda x: x.rstrip())

    output = output[str_char_len:]
    for action in actions:
        output = action(output)
    return str(parse_escape(output))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['modcfg']

package_data = \
{'': ['*']}

install_requires = \
['lark-parser>=0.11.2,<0.12.0']

setup_kwargs = {
    'name': 'modcfg',
    'version': '0.1.2',
    'description': 'Yes, another config lang',
    'long_description': '# ModCFG\n![YEs](https://raw.githubusercontent.com/ThatXliner/modcfg/master/noo.png)\n\n[![codecov](https://codecov.io/gh/ThatXliner/modcfg/branch/master/graph/badge.svg)](https://codecov.io/gh/ThatXliner/modcfg) [![Documentation Status](https://readthedocs.org/projects/modcfg/badge/?version=latest)](https://modcfg.readthedocs.io/en/latest/?badge=latest) [![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black) [![CI](https://github.com/ThatXliner/modcfg/actions/workflows/ci.yml/badge.svg?branch=master)](https://github.com/ThatXliner/modcfg/actions/workflows/ci.yml)\n\n> YEs, another configuration language\n\nSeriously, how many configuration languages do we have? From the top of my head, I can only name:\n\n - [JSON](https://www.json.org/)\n - [YAML](https://yaml.org)\n - [INI](https://wikipedia.org/wiki/INI_file)\n - [TOML](https://toml.io/)\n - [XML](https://www.w3.org/XML/)\n\nHmmm... that\'s *it* I think. wE NeEd MOrE cOnfIGURaTion LanGuaGEs!!1!1!!\n\n## But actually though\n\nI originally based this off of the Ruby/DSL for homebrew formulas.\n\nHere are my complaints for the alternatives:\n\n - JSON, while simple, isn\'t too readable and **only has the bear minimum of datatypes.** That may be a feature for some people, though.\n - YAML is [**dangerous**](https://www.arp242.net/yaml-config.html#insecure-by-default). And [**unpredictable**](https://hitchdev.com/strictyaml/why/implicit-typing-removed/).\n - INI and TOML are quite similar and they both are quite nice with a lot of datatypes but, in my opinion, **the format isn\'t as natural to read as YAML**. TOML supports dates and datetime, though.\n - Finally, XML is garbage. It\'s barely human-readable and can decode into a super wonky format. So no.\n\nFinally, I just wanted an excuse to play with Lark.\n\n### Features\n\n - Enums with a built-in resolver\n - Date time and Date support\n - YAML-like syntax\n - Support for comments, **unicode escape sequences**, **string modifications**, and **cleanly managed indented strings**\n\n### Examples\n\nGiven this modcfg document:\n```yaml\nmodule hello_world:\n    hello => world\n    this: "also works"\n    \'single quotes\' = "equals double quotes"\n    how -> {\n            about: {\n                some:\n                    - very\n                    - crazy\n                    - data:\n                        structures = o_0\n            }\n        }\n```\nand this python script:\n```py\nimport modcfg\nmodcfg.loads(DOC)  # `DOC` is the document above\n```\nThe output is\n```py\n[\n    Module(\n        name="hello_world",\n        contents=[\n            {\n                "hello": "world",\n                "this": "also works",\n                "single quotes": "equals double quotes",\n                "how": {\n                    "about": {\n                        "some": ["very", "crazy", {"data": {"structures": "o_0"}}]\n                    }\n                },\n            }\n        ],\n    )\n]\n```\nCrazy, right? It gets better with enums and date(time)s... You might as well read the whole [documentation][documentation].\n\n## Installation\n\nI have not uploaded this to the PyPi. You must develop-install instead.\n\nThe classic method:\n\n```bash\n$ git clone https://github.com/ThatXliner/modcfg.git\n$ cd modcfg\n$ python3 -m venv .venv\n$ source .venv/bin/activate\n$ pip install -e .\n```\n\nThe best method (requires [Poetry](https://python-poetry.org/)):\n\n```bash\n$ git clone https://github.com/ThatXliner/modcfg.git\n$ cd modcfg\n$ poetry install\n```\n\n## FAQ\n\n### Why did you make this\n\n![why](https://raw.githubusercontent.com/ThatXliner/modcfg/master/why.png)\n\n[Lark](https://github.com/lark-parser/lark) is epic.\n\n### Why the name?\n\n**Mod**ular **C**on**f**i**g**uration language.\n\n[documentation]: https://modcfg.readthedocs.io/en/latest/index.html\n',
    'author': 'Bryan Hu',
    'author_email': 'bryan.hu.2020@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ThatXliner/modcfg',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
